import os
import random
import numpy as np
import rasterio as rio
from rasterio.windows import Window
from rasterio.enums import Resampling
from affine import Affine
from imagebox.config import FIRST, LAST, BAND_ORDERING
from . import utils
#
# CONSTANTS
#
RESAMPLING=Resampling.bilinear

#
# READ/WRITE
#
def read(
        path,
        window=None,
        window_profile=True,
        return_profile=True,
        res=None,
        scale=None,
        out_shape=None,
        bands=None,
        resampling=RESAMPLING,
        band_ordering=None,
        dtype=None):
    """ read image
    Args: 
        - path<str>: source path
        - window<tuple|Window>: col_off, row_off, width, height
        - window_profile<bool>:
            - if True return profile for the window data
            - else return profile for the src-image
        - res<int>: rescale to new resolution. overides scale and out_shape
        - scale<float>: rescale image res=>res*scale overrides out_shape
        - out_shape<tuple>: (h,w) rescales image. overwritten by res and scale
        - dtype<str>:
    Returns:
        <tuple> np.array, image-profile
    """
    with rio.open(path,'r') as src:
        if return_profile:
            profile=src.profile
        if window:
            w,h=window[2], window[3]
            window=Window(*window)
            if window_profile and return_profile:
                profile['transform']=src.window_transform(window)
                profile['width']=w
                profile['height']=h
        else:
            w,h=src.width, src.height
        if res:
            scale=src.res[0]/res
        if scale:
            out_shape=(int(h*scale),int(w*scale))
        if out_shape and return_profile:
            profile=rescale_profile(profile,out_shape)
        image=src.read(
                indexes=bands,
                window=window,
                out_shape=out_shape,
                resampling=resampling )
        if dtype:
            image=image.astype(dtype)
        image=utils.order_bands(image,band_ordering)
    if return_profile:
        return image, profile
    else:
        return image


def write(im,path,profile,makedirs=True):
    """ write image
    Args: 
        - im<np.array>: image
        - path<str>: destination path
        - profile<dict>: image profile
        - makedirs<bool>: if True create necessary directories
    """  
    if makedirs:
        dirname=os.path.dirname(path)
        if dirname:
            os.makedirs(os.path.dirname(path),exist_ok=True)
    affine_transform=profile.get('affine')
    if affine_transform:
        profile['transform']=affine_transform
    with rio.open(path,'w',**profile) as dst:
        dst.write(im)
        

def read_stack(paths,res_list=None,stack_res=FIRST,resampling=RESAMPLING):
    """ band-wise read for images

    Args: 
        - paths<list>: list of source paths (per-band/ordered)
        - res_list<list|None>: 
            * list of resolutions.  
            * if passed will rescale bands to stack_res
        - stack_res:
            * resolution to rescale all bands to
            * if 'first' use the first resolution in res_list
        - resampling<str>: resampling method
    """
    with rio.open(paths[0]) as src:
        profile = src.profile
    profile.update(count=len(paths))
    if res_list:
        if isinstance(res_list,int):
            res_list=[res_list]*len(paths)
        if stack_res is FIRST:
            stack_res=res_list[0]
        scale=(stack_res/res_list[0])
        if scale!=1.0:
            profile['width']*=scale
            profile['height']*=scale
        ims=[ read(p,scale=r/stack_res,band=1,resampling=resampling) 
              for p,r in zip(paths,res_list) ]
    else:
        ims=[read(p,return_profile=False) for p in paths]

    return np.concatenate(ims), profile        


#
# WINDOW/PROFILE HELPERS
#
def rescale_profile(profile,out_shape):
    affine=profile['transform']
    h_out,w_out=out_shape
    h,w=profile['height'],profile['width']
    res_y=int(affine.e*h/h_out)
    res_x=int(affine.a*w/w_out)
    profile['transform']=Affine(res_x, 0.0, affine.c,0.0, res_y, affine.f)
    profile['height'],profile['width']=h_out,w_out
    return profile










import random
import numpy as np
from scipy.signal import convolve2d
from imagebox.config import FIRST, LAST, BAND_ORDERING, BANDS_FIRST
#
# CONSTANTS
#
IMAGE='image'
DEFAULT_VMAP_VALUE=IMAGE
BANDS_FIRST_AXES=(1,2)
BANDS_LAST_AXES=(0,1)
DENORM_ERROR='imagebox.processor.denormalize: bands last not yet implemented'
SWAP_BANDS_ERROR='imagebox.processor._swap_bands_axes: im.ndim must be 3 or 4'
SMOOTHING_KERNEL=np.ones((3,3))


#
# METHODS
#
def center(im,means=None,to_int=False,bands_first=BANDS_FIRST):
    """ center image array
    Args:
        im<np.array>: image array
        means<np.array|list>: 
            - band-wise means to center the array around
            - if none use the band-wise means of the image itself
        to_int<bool>: if true convert to uint8 after centering
        bands_first<bool>: true if array is bands first
    """
    if means is None:
        means=np.mean(im,axis=_axes(im.ndim,bands_first))
    if bands_first:
        means=_to_vector(means)
    im=(im-means)
    if to_int:
        im=im.round().astype(np.uint8)
    return im


def normalize(im,means=None,stdevs=None,bands_first=BANDS_FIRST):
    """ normalize image array
    Args:
        im<np.array>: image array
        means<np.array|list>: 
            - band-wise means to center the array around
            - if none use the band-wise means of the image itself
        stdevs<np.array|list>: 
            - band-wise standard deviations
            - if none use the band-wise standard deviations of the image itself
        bands_first<bool>: true if array is bands first
    """
    if stdevs is None:
        stdevs=np.std(im,axis=_axes(im.ndim,bands_first))   
    im=center(im,means=means,to_int=False,bands_first=bands_first)
    if bands_first:
        stdevs=_to_vector(stdevs)
    return im/stdevs


def denormalize(im,means,stdevs,bands=[0,1,2],bands_first=BANDS_FIRST,dtype=np.uint8):
    """ denormalize image array
    Args:
        im<np.array>: image array
        means<np.array|list>: means
        stdevs<np.array|list>: stdevs
        bands_first<bool>: true if array is bands first
    """ 
    if not bands_first:
        im=im[:,:,bands]
        im=im.transpose(2,0,1)
    else:
        im=im[bands]
    stdevs=np.array(stdevs)[bands]
    means=np.array(means)[bands]
    stdevs=stdevs.reshape((im.shape[0],1,1))
    means=means.reshape((im.shape[0],1,1))
    im=stdevs*im+means
    if not bands_first:
        im=im.transpose(1,2,0)
    return im.astype(dtype)


def map_values(im,value_map,default_value=DEFAULT_VMAP_VALUE):
    """ map values of image array
    Args:
        im<np.array>: image array
        value_map<dict>: 
            - keys: new values in mapped_im
            - values<list>: list of values to be mapped to corresponding key  
        default_value<int|float|np.nan|'image'>: 
            if default_value is 'image' use image values for unmapped values
            else use default value for unmapped values
    """
    value_map=value_map.copy()
    default_value=value_map.pop('.default',default_value)
    if default_value==IMAGE:
        mapped_im=np.array(im).copy()
    else:
        mapped_im=np.full_like(im,default_value)
    for k,v in value_map.items():
        mapped_im[np.isin(im,v)]=k
    return mapped_im


def to_categorical(im,nb_categories):
    """ to categorical
    map single band int valued images to multi-band binary value image
    """
    return np.eye(nb_categories)[:,im]



def categorical_smoothing(im,nb_categories,kernel=SMOOTHING_KERNEL):
    """ smooth categorical inputs
    """
    im=to_categorical(im,nb_categories)
    for i in range(nb_categories):
        im[i]=convolve2d(im[i],kernel,mode='same')
    return im.argmax(axis=0)


def crop(im,cropping,bands_first=BANDS_FIRST):
    """ crop image """
    if bands_first:
        if im.ndim==4:
            return im[:,:,cropping:-cropping,cropping:-cropping]
        elif im.ndim==3:
            return im[:,cropping:-cropping,cropping:-cropping]    
        elif im.ndim==2:
            return im[cropping:-cropping,cropping:-cropping]
        elif im.ndim==1:
            return im[cropping:-cropping]
    else:
        if im.ndim>=2:
            return im[cropping:-cropping,cropping:-cropping]
        else:
            return im[cropping:-cropping]



def pad(im,padding=1,axes=None,value=0):
    """ pad image along axes
    * im<np.array>: image to pad
    * axes<None|int|tuple>: 
        - axis or axes to pad
        - if None assume bands last
    * padding<int|tuple>: number of pixes to pad by. tuple for have asymmetric padding
    * value<number>: constant value to pad with
    """
    if isinstance(axes,int):
        axes=(axes)
    elif axes is None:
        if im.ndim==4:
            axes=(2,3)
        elif im.ndim==3:
            axes=(1,2)
        elif im.ndim==2:
            axes=(0,1)
        elif im.ndim==1:
            axes=(0)
    def pad_tupl(axis,pad_tupl):
        if axis in axes:
            return pad_tupl
        else:
            return (0,0)
    if isinstance(padding,int):
        width=(padding,padding)
    else:
        width=padding
    return np.pad(
        im,
        pad_width=[pad_tupl(im,width) for im in range(im.ndim)],
        mode='constant',
        constant_values=value)


def augmentation(k=None,flip=None):
    """ get k(rotation), flip values for image augmentation

    For both k and flip passing None will randomly select a value
    for k and flip. otherwise the k, flip value will be returned

    Args:
        k<int|None|False>: number of 90 degree rotations
        flip<bool|None>: true/false to flip or not flip
    """
    if k is not False: 
        if not k: k=random.randint(0,3)
    if flip is not False:
        if not flip: flip=random.choice([True,False])
    return k,flip


def augment(im,k=False,flip=False,bands_first=BANDS_FIRST,random=False):
    """ augment (rotate/flip) image

    Args:
        im<np.array>: image array
        k<int|False>: number of 90 degree rotations 
        flip<bool>: flip or don't flip
        random<bool>: if true get augmentation first
    """
    if random:
        k, flip=augmentation()
    if k is not False:
        im=rotate(im,k,bands_first=bands_first)
    if flip is not False:
        im=flip_image(im,bands_first=bands_first)
    return im


def rotate(im,k,bands_first=BANDS_FIRST):
    """ rotate image
    Args:
        im<np.array>: image array
        k<int|False>: number of 90 degree rotations

    * PYTORCH HACK:
    * - im=im+0
    * - negative stride issue 
    * - https://discuss.pytorch.org/t/torch-from-numpy-not-support-negative-strides/3663/7
    """
    im=np.rot90(im,k,axes=_axes(im.ndim,bands_first))
    im=im+0
    return im


def flip_image(im,bands_first=BANDS_FIRST,axis=None):
    """ flip image
    Args:
        im<np.array>: image array

    * PYTORCH HACK:
    * - im=im+0
    * - negative stride issue 
    * - https://discuss.pytorch.org/t/torch-from-numpy-not-support-negative-strides/3663/7
    """
    if axis is None:
        if (im.ndim==2) or (not bands_first):
            axis=0
        else:
            axis=1
    im=np.flip(im,axis)
    im=im+0
    return im


#
# HELPERS
#
def rgb_rescale(
        im,
        bands=None,
        rgb_max=255,
        im_max=2500,
        dtype=np.uint8,
        bands_first=BANDS_FIRST):
    if bands_first:
        if bands:
            im=im[bands]
        else:
            im=im[:3]
    else:
        if im.ndim==4:
            if bands:
                im=im[:,:,:,bands]
            else:
                im=im[:,:,:,:3]
        else:
            if bands:
                im=im[:,:,bands]
            else:
                im=im[:,:,:3]
    if im_max:
        im=im.astype(np.float)*rgb_max/im_max
    im=im.clip(0,rgb_max)
    if dtype:
        im=im.astype(dtype)
    return im


def is_bands_first(im):
    """ checks if image is bands last or bands first

    Note: Assumes the number of bands is less than or equal to the 
    width of the image.

    Returns: True/False
    """
    shape=im.shape
    return shape[-3]<=shape[-1]


def to_bands_last(im):
    """ convert image to bands last
    """
    shape=im.shape
    if (im.ndim>2) and is_bands_first(im):
        im=_swap_bands_axes(im)
    return im


def to_bands_first(im):
    """ convert image to bands first
    """    
    shape=im.shape
    if (im.ndim>2) and not is_bands_first(im):
        im=_swap_bands_axes(im)
    return im



#
# INTERNAL
#
def _axes(ndim,bands_first):
    if (ndim==2) or (not bands_first):
        return BANDS_LAST_AXES
    else:
        return BANDS_FIRST_AXES



def _swap_bands_axes(im):
    ndim=im.ndim
    if ndim==3:
        im=im.transpose(1,2,0)
    elif ndim==4:
        im=im.transpose(0,2,3,1)
    else:
        raise ValueError(SWAP_BANDS_ERROR)
    return im


def _to_vector(arr):
    return np.array(arr).reshape(-1,1,1)




import numpy as np
from imagebox.config import FIRST, LAST, BAND_ORDERING, BANDS_FIRST
#
# CONSTANTS
# 
EPS=1e-8


#****************************************************************
#
# IMPORTANT NOTE: band index definition based on band order
#                 red, green, blue, nir, red-edge, swir1
#
#****************************************************************

BANDS=[
    'red',
    'green',
    'blue',
    'nir',
    'red-edge',
    'swir1',
    'swir2'
]
INDICES={
    'ndvi':(3,0),
    'ndwi':(1,3),
    'ndwi_leaves':(3,5),
    'ndbi':(5,3),
    'built_up': ['ndbi','ndvi'],
    'greeness':{ 
        "numerator_bands":1,
        "denominator_bands":0,
    },
   'chlogreen':{
        "numerator_bands":3,
        "denominator_bands":[1,4],
    },
    'gcvi':{
        "numerator_bands":3,
        "denominator_bands":1,
        "constant":1
    },
    'evi_modis':{
        "numerator_bands":[3,0],
        "numerator_coefs":[2.5,-2.5],
        "denominator_bands":[3,0,2],
        "denominator_coefs":[1,6,7.5],
        "denominator_constant":1
    },
    'evi_s2':{
        "numerator_bands":[3,0],
        "numerator_coefs":[2.5,-2.5],
        "denominator_bands":[3,0,2],
        "denominator_coefs":[1,6,7.5],
        "denominator_constant":10000
    }
}


ORDERED='ordered'
ORDERD_BANDS=[
    'coastal-aerosol',
    'blue',
    'green',
    'red',
    'red-edge',
    'red-edge-2',
    'red-edge-3',
    'nir',
    'red-edge-4',
    'water-vapor',
    'cirrus',
    'swir1',
    'swir2'
]
INDICES_ORDERED={
    'ndvi':(7,3),
    'ndwi':(2,7),
    'ndwi_leaves':(7,11),
    'ndbi':(11,7),
    'built_up': ['ndbi','ndvi'],
}

S2_1020='s2_1020'
S2_1020_BANDS=[
    'B2',
    'B3',
    'B4',
    'B5',
    'B6',
    'B7',
    'B8',
    'B11',
    'B12'
]
INDICES_S2_1020={
    'ndvi':(6,2),
    'ndwi':(1,6),
    'ndwi_leaves':(6,8),
    'ndbi':(8,6),
    'built_up': ['ndbi','ndvi'],
}



LSAT_SR='lsat_sr'
LSAT_SR_BANDS=[
    'B1', #Band 1 (ultra blue) surface reflectance     0.435-0.451 μm  0.0001
    'B2', #Band 2 (blue) surface reflectance       0.452-0.512 μm  0.0001
    'B3', #Band 3 (green) surface reflectance      0.533-0.590 μm  0.0001
    'B4', #Band 4 (red) surface reflectance        0.636-0.673 μm  0.0001
    'B5', #Band 5 (near infrared) surface reflectance      0.851-0.879 μm  0.0001
    'B6', #Band 6 (shortwave infrared 1) surface reflectance       1.566-1.651 μm  0.0001
    'B7', #Band 7 (shortwave infrared 2) surface reflectance       2.107-2.294 μm  0.0001
    'B10', #Band 10 brightness temperature. This band, while originally collected with a resolution of 100m / pixel, has been resampled using cubic convolution to 30m. Kelvin  10.60-11.19 μm  0.1
    'B11', #Band 11 brightness temperature. This band, while originally collected with a resolution of 100m / pixel, has been resampled using cubic convolution to 30m. Kelvin  11.50-12.51 μm  0.1
]
INDICES_LSAT_SR={
    'ndvi':(4,3),
    'ndwi':(2,4),
    'ndwi_leaves':(4,5),
    'ndbi':(5,4),
    'built_up': ['ndbi','ndvi'],   
}



#
# METHODS
#
def index(im,index_name,indices=None):
    """ band index based on name, or ndiff args or ratio-index args
    
    Args:
        im<np.array>: image array
        index_name<str|False>: index key from INDICES or False for custom index
    """
    if isinstance(indices,str):
        if indices==ORDERED:
            indices=INDICES_ORDERED
        elif indices==S2_1020:
            indices=INDICES_S2_1020
        else:
            indices=INDICES
    elif not indices: 
        indices=INDICES
    args=indices[index_name]
    if isinstance(args,dict):
        return ratio_index(im,**args)
    else:
        if isinstance(args[0],str):
            a=index(im,args[0])
            b=index(im,args[1])
            return a-b
        else:
            return normalized_difference(im,*args)


def normalized_difference(im,band_1,band_2,bands_first=BANDS_FIRST):
    """ Normalize Difference

        Args:
            im<np.array>: image array
            b1<int>: the band number for b1
            b2<int>: the band number for b2  


        Returns:
            <arr>: (b1-b2)/(b1+b2)
    """
    im=im.astype(np.float)
    if bands_first:
        band_1=im[band_1]
        band_2=im[band_2]
    else:
        band_1=im[:,:,band_1]
        band_2=im[:,:,band_2]
    return np.divide(band_1-band_2,band_1+band_2+EPS)


def ratio_index(
        im,
        numerator_bands,
        denominator_bands=None,
        numerator_coefs=None,
        denominator_coefs=None,
        numerator_constant=0,
        denominator_constant=0,
        constant=0):
    """ Ratio Index
        
        Generalized Index that allows for any linear combination of bands
        in numerator and denominator plus an overall constant.

    """
    if not constant:
        constant=0
    numerator=linear_combo(
        im,
        bands=numerator_bands,
        coefs=numerator_coefs,
        constant=numerator_constant)
    if denominator_bands is None:
        denominator=1
    else:  
        denominator=linear_combo(
            im,
            bands=denominator_bands,
            coefs=denominator_coefs,
            constant=denominator_constant)
    return np.divide(numerator,denominator+EPS)+constant


def linear_combo(im,bands,coefs=None,constant=None,bands_first=BANDS_FIRST):
    """ Linear Combination

        Args:
            im<np.array>: image array   
            bands <list|int>: list of band indices or band index
            coefs <list|int|None>: list of coefs, a coef for all bands, or None -> 1
            constant <float[0]>: additive constant
        Returns:
            <np.array>: image_bands dot coefs + constant
    """
    im=im.astype(np.float)
    if not constant:
        constant=0
    if isinstance(bands,int):
        bands=[bands]
    if not coefs:
        coefs=1
    if isinstance(coefs,int):
        coefs=[coefs]*len(bands)
    if bands_first:
        im=coefs[0]*im[bands[0]]
        for c,b in zip(coefs[1:],bands[1:]):
            im+=c*im[b]
    else:
        im=coefs[0]*im[:,:,bands[0]]
        for c,b in zip(coefs[1:],bands[1:]):
            im+=c*im[:,:,b]
    return im+constant


def shadow_mask(im,band_bounds=[77,77,87],max_diff=25,bands=[0,1,2],blueness=6):
    bbnds=[]
    for b,bnd in zip(bands,band_bounds):
        bbnds.append(im[b]<=bnd)
    bbnds=np.array(bbnds)
    isblue=im[bands[:-1]].max(axis=0)+blueness<=im[bands[-1]]
    isgrey=(im[bands[-1]]-im[bands[:-1]].min(axis=0))<=max_diff
    return bbnds.all(axis=0)*isblue*isgrey




import rasterio.transform as transform
from rasterio.crs import CRS
from imagebox.config import FIRST, LAST, BAND_ORDERING



def get_crs(crs,as_dict=False):
    if isinstance(crs,int):
        crs_dict={'init':f'epsg:{crs}'}
    elif isinstance(crs,str):
        crs_dict={'init':crs}
    else:
        return crs
    if as_dict:
        return crs_dict
    else:
        return CRS(crs_dict)


def crs_res_bounds(profile):
    """ get crs, resolution and bounds form image profile """
    affine=profile['transform']
    res=affine.a
    minx=affine.c
    miny=affine.f-profile['height']*res
    maxx=minx+profile['width']*res
    maxy=miny+profile['height']*res
    crs=str(profile['crs'])
    return crs,res,(minx,miny,maxx,maxy)



def bounds_from_profile(profile):
    """ bounds from profile """
    a=profile['transform']
    h=profile['height']
    w=profile['width']
    return transform.array_bounds(h,w,a)


def profile(
        crs,
        transform,
        width=None,
        height=None,
        size=None,
        count=1,
        nodata=None,
        dtype='uint8',
        compress='lzw'):
    """ construct profile """
    if size:
        width=height=size
    return {
        'crs': get_crs(crs),
        'transform': transform,
        'width': width,
        'height': height,
        'count': count,
        'nodata': nodata,
        'dtype': dtype,
        'compress': compress,
        'driver': 'GTiff',
        'interleave': 'pixel' }


#
# IO
#
def order_bands(image,band_ordering=None):
    if band_ordering is None:
        band_ordering=BAND_ORDERING 
    if band_ordering.lower()==LAST:
        image=image.transpose(1,2,0)
    return image


#
# WINDOW HELPERS
#
def window_origin(src_transform,target_transform):
    """ origin of window from src and window affine-transform """
    x=(target_transform.c-src_transform.c)/target_transform.a
    y=(target_transform.f-src_transform.f)/target_transform.e
    return round(x),round(y)


def profiles_to_window(src_profile,target_profile):
    """ pixel window from src and window profile """
    x,y=window_origin(src_profile['transform'],target_profile['transform'])
    return Window(x,y,target_profile['width'],target_profile['height'])


def crop_src_by_target_profile(src,target_profile):
    im=src.read()
    window=profiles_to_window(src.profile,target_profile)
    imin=window.row_off
    jmin=window.col_off
    imax=imin+window.width
    jmax=jmin+window.height
    return im[:,imin:imax,jmin:jmax]






import os
import yaml
#
# CONSTANTS
#
IMAGE_BOX_CONFIG_FILE=f'{os.getcwd()}/imagebox.config.yaml'
FIRST='first'
LAST='last'
NOISY=os.environ.get('IMAGE_BOX_NOISE',False)


#
# CONFIG
#
try: 
    with open(IMAGE_BOX_CONFIG_FILE,'rb') as file:
        _config=yaml.safe_load(file)
        if NOISY:
            print(f'IMAGE_BOX: config file loaded ({IMAGE_BOX_CONFIG_FILE})')
except Exception as e:
    _config={}
    if NOISY:
        print(f'IMAGE_BOX: config file ({IMAGE_BOX_CONFIG_FILE}) not found')


BAND_ORDERING=_config.get(
    'band_ordering',
    os.environ.get('IMAGE_BOX_BAND_ORDERING',FIRST))


#
# DERIVED
#
BANDS_FIRST=BAND_ORDERING==FIRST



from distutils.core import setup
setup(
  name = 'imagebox',
  packages = ['imagebox'],
  version = '0.0.0.21',
  description = 'ImageBox: python utilities for working with multispectral imagery',
  author = 'Brookie Guzder-Williams',
  author_email = 'brook.williams@gmail.com',
  url = 'https://github.com/brookisme/imagebox',
  download_url = 'https://github.com/brookisme/imagebox/tarball/0.1',
  keywords = ['python','geotiff','rasterio','image','io'],
  include_package_data=True,
  data_files=[
    (
      'config',[]
    )
  ],
  install_requires=[
    'numpy',
    'rasterio',
    'pyyaml',
    'affine',
    'rasterio',
    'gcs_helpers>=0.0.0.22'
  ],
  classifiers = [],
  entry_points={
      'console_scripts': [
      ]
  }
)


import math
from random import randint
from rasterio.enums import Resampling
import numpy as np
import gcs_helpers.fetch as gfetch
import imagebox.io as io
import imagebox.processor as proc
import imagebox.indices as indices
from imagebox.config import FIRST, LAST, BAND_ORDERING, BANDS_FIRST

#
# CONSTANTS
# 
INPUT_DTYPE=np.float
TARGET_DTYPE=np.int64
DEFAULT_SIZE=256
DEFAULT_OVERLAP=0
INPUT_RESAMPLING=Resampling.bilinear
TARGET_RESAMPLING=Resampling.mode
TO_CATEGORICAL_ERROR=(
    'imagebox.handler: '
    'nb_categories required for to_categorical' )
SAFE_RESCALE_ERROR=(
    'imagebox.handler: '
    'target rescale leads to fractional value. '
    'use safe_rescale=False to force' )
DIMS_REQUIRED_ERROR=(
    'imagebox.handler: '
    'width and height, or example_path, required for (float)cropping' )

#
# InputTargetHandler
#
class InputTargetHandler(object):
    """ 
    
    A handler for processing multiple input/target pairs in the 
    same way. 

    This is especially useful for constructing data loader/generators

    Args:
        input_bands<list|None>: list of bands to select from image
        means,stdevs<list|np.array>:
            - means/stdevs for all bands in input image (before selecting input_bands)
            - if means and not stdevs: center input around means
            - if means and stdevs normalize inputs
        band_indices<list>:
            - list of arguments for imagebox.indices.index
            - adds band_index bands to image
        indices_dict<dict>:
            - preset indices from which to select band_indices
        value_map<dict>: 
            - target value map. 
            - keys: new values
            - values: list of values to map to new value 
        default_mapped_value<'image'|int|float>:
            - only used if value_name is not None
            - if 'image' keep all unmapped values
            - otherwise value for all unmapped values
        to_categorical<bool>: one-hot target image 
        nb_categories<int>: 
            - number of target categories
            - required for to_categorical
        augment<bool>: augment image
        flip_target/input<bool>: 
            - flip target/input image before any additional augmentation 
            - for use when target and input pairs have opposite y-axis orientation
            - set only one of these to be true
        target/input_resolution: rescale input or target with input/target resampling method
        input/target_resampling: resampling method for input or target
        padding|input/target_padding<int|None>: 
            - amount to pad images|input/target-image after processing the image
            - will be ignored if there is cropping
        cropping<int|None>:
            - amount to crop both input and target image when reading the image
        input/target_cropping<int|None>: 
            - amount to crop input/target image after processing the image
        float_cropping<int|float|None>: 
            - remove pixels from h/w of input and target starting at random i,j.
            - 2*float_cropping must be int
        width<int|None>: image width (required if float cropping and not tiller)
        height<int|None>: image height (required if float cropping and not tiller)
        tiller<Tiller|True|None>:
            tile the image then read in a specific tile
        tiller_config<dict>:
            if tiller is True create tiller=Tiller(**tiller_config)
        input_dtype<str>: input data type
        target_dtype<str>: target data type

    """ 
    def __init__(self,
            input_bands=None,
            means=None,
            stdevs=None,
            band_indices=None,
            indices_dict=None,
            value_map=None,
            default_mapped_value=proc.DEFAULT_VMAP_VALUE,
            to_categorical=False,
            nb_categories=None,
            augment=True,
            flip_target=False,
            flip_input=False,
            input_resolution=None,
            target_resolution=None,
            input_bounds=None,
            input_resampling=INPUT_RESAMPLING,
            target_resampling=TARGET_RESAMPLING,
            padding=None,
            input_padding=None,
            target_padding=None,
            input_padding_value=0,
            target_padding_value=0,
            stop_floating=False,
            float_cropping=None,           
            target_ratio=1,
            safe_rescale=True,
            cropping=None,
            input_cropping=None,
            target_cropping='auto',
            window_index=None,
            size=None,
            width=None,
            height=None,
            example_path=None,
            tiller=None,
            tiller_config={},
            target_expand_axis=None,
            input_preprocess=None,
            target_preprocess=None,
            target_squeeze=True,
            read_from_gcs=False,
            input_dtype=INPUT_DTYPE,
            target_dtype=TARGET_DTYPE ):
        if tiller is True:
            self.tiller=Tiller(**tiller_config)
        else:
            self.tiller=tiller
        self.input_bands=input_bands
        self.band_indices=band_indices
        self.indices_dict=indices_dict
        self.value_map=value_map
        self.list_value_map=isinstance(value_map,list)
        self.default_mapped_value=default_mapped_value
        self.means=means
        self.stdevs=stdevs
        if to_categorical and (not nb_categories):
            raise ValueError(TO_CATEGORICAL_ERROR)
        self.to_categorical=to_categorical
        self.nb_categories=nb_categories
        self.augment=augment
        self.flip_target=flip_target
        self.flip_input=flip_input
        self.input_resolution=input_resolution
        self.target_resolution=target_resolution
        self.input_bounds=input_bounds
        self.input_resampling=input_resampling
        self.target_resampling=target_resampling
        self.set_augmentation()
        self.input_padding=input_padding or padding
        self.target_padding=target_padding or padding
        self.input_padding_value=input_padding_value
        self.target_padding_value=target_padding_value
        self._set_cropping_and_dims(
            cropping,
            target_ratio,
            safe_rescale,
            input_cropping,
            target_cropping,
            float_cropping,
            stop_floating,
            size,
            width,
            height,
            example_path)
        self.set_window(window_index=window_index)
        self.target_expand_axis=target_expand_axis
        self.input_preprocess=input_preprocess
        self.target_preprocess=target_preprocess
        self.target_squeeze=target_squeeze
        self.read_from_gcs=read_from_gcs
        self.input_dtype=input_dtype
        self.target_dtype=target_dtype


    def input(self,path,window=None,means=None,stdevs=None,return_profile=False):
        self.input_path=path
        im,profile=self._read(
            path,
            self.input_resolution,
            self.target_resampling,
            window or self.input_window )
        if means is None:
            means=self.means
        if stdevs is None:
            stdevs=self.stdevs

        im=process_input(
            im,
            preprocess=self.input_preprocess,
            flip=self.flip_input,
            input_bands=self.input_bands,
            band_indices=self.band_indices,
            indices_dict=self.indices_dict,
            padding=self.input_padding,
            padding_value=self.input_padding_value,
            bounds=self.input_bounds,
            means=means,
            stdevs=stdevs,
            dtype=self.input_dtype )
        return self._return_data(
            im,
            profile,
            return_profile )


    def target(self,path,window=None,return_profile=False):
        self.target_path=path
        im,profile=self._read(
            path,
            self.target_resolution,
            self.target_resampling,
            window or self.target_window )
        im=process_target(
            im,
            preprocess=self.target_preprocess,
            flip=self.flip_target,
            value_map=self.value_map,
            list_value_map=self.list_value_map,
            default_mapped_value=self.default_mapped_value,
            categorical=self.to_categorical,
            nb_categories=self.nb_categories,
            padding=self.target_padding,
            padding_value=self.target_padding_value,
            expand_axis=self.target_expand_axis,
            squeeze=self.target_squeeze,
            dtype=self.target_dtype )
        if self.value_map and self.list_value_map:
            return [self._return_data(o,profile,False) for o in im]
        else:
            return self._return_data(im,profile,return_profile)


    def set_augmentation(self,k=None,flip=None):
        if self.augment:
            self.k, self.flip=proc.augmentation(k,flip)
        else:
            self.k=False
            self.flip=False
    

    def set_window(self,window=None,window_index=None,example_path=None):
        if window:
            input_window=window
            target_window=window
        elif self.tiller:
            if window_index is None:
                window_index=randint(0,len(self.tiller)-1)
            self.window_index=window_index            
            target_window=self.tiller[self.window_index]
            input_window=target_window
        else:
            self.window_index=False
            input_window=0,0,self.input_width,self.input_height
            target_window=0,0,self.target_width,self.target_height
        self.input_window=self._shift_crop_window(
                input_window,
                dx=self.input_cropping,
                dy=self.input_cropping,
                crop=self.input_cropping)
        self.target_window=self._shift_crop_window(
                target_window,
                dx=self.target_cropping,
                dy=self.target_cropping,
                crop=self.target_cropping)
        if self.float_cropping:
            self.float_x=self._random_delta()
            self.float_y=self._random_delta()
            self.input_window=self._shift_crop_window(
                    self.input_window,
                    dx=self.float_x,
                    dy=self.float_y,
                    crop=self.float_cropping )
            self.target_window=self._shift_crop_window(
                    self.target_window,
                    dx=self._target_rescale(self.float_x),
                    dy=self._target_rescale(self.float_y),
                    crop=self._target_rescale(self.float_cropping) )


    #
    # INTERNAL METHODS
    #
    def _set_cropping_and_dims(self,
            cropping,
            target_ratio,
            safe_rescale,
            input_cropping,
            target_cropping,
            float_cropping,
            stop_floating,
            size,
            width,
            height,
            example_path):
        self.target_cropping=None
        self.float_cropping=None
        self.target_ratio=target_ratio or 1
        self.input_cropping=input_cropping or cropping or 0
        float_cropping=float_cropping or 0
        if not float(2*float_cropping).is_integer():
            raise ValueError('`2*float_cropping` must be int')
        self.safe_rescale=safe_rescale
        if size:
            self.input_width=size
            self.input_height=size
        else:
            self.input_width=width
            self.input_height=height
        self._ensure_dimensions(example_path)
        self.target_width=self._target_rescale(self.input_width)
        self.target_height=self._target_rescale(self.input_height)
        if stop_floating:
            self.input_cropping+=float_cropping
            self.float_cropping=False
        else:
            self.float_cropping=float_cropping
        if self.input_cropping and (target_cropping=='auto'):
            self.target_cropping=self._target_rescale(self.input_cropping)
        elif isinstance(target_cropping,(int,float)):
            self.target_cropping=target_cropping
        else:
            self.target_cropping=cropping or 0


    def _target_rescale(self,value):
        if self.target_ratio==1:
            return value
        else:
            fval=value*self.target_ratio
            val=round(fval)
            if (not self.safe_rescale) or (val==fval):
                return val
            else:
                raise ValueError(SAFE_RESCALE_ERROR)


    def _read(self,path,resolution,resampling,window):
        if self.read_from_gcs:
            im,p=gfetch.image(
                path=path,
                window=window,
                res=resolution,
                resampling=resampling,
                band_ordering=BAND_ORDERING)
        else:
            im,p=io.read(
                path,
                window=window,
                res=resolution,
                resampling=resampling)
        return im,p


    def _random_delta(self):
        return randint(0,2*int(self.float_cropping*self.target_ratio))


    def _shift_crop_window(self,window,dx=0,dy=0,crop=0):
        if dx or dy or crop:
            x=window[0]+dx
            y=window[1]+dy
            w=window[2]-2*crop
            h=window[3]-2*crop
            return int(x),int(y),int(w),int(h)
        else:
            return window


    def _ensure_dimensions(self,example_path):
        if not (self.input_width and self.input_height):
            if example_path:
                shape=io.read(example_path,return_profile=False).shape
                if BANDS_FIRST:
                    self.input_height,self.input_width=shape[1:]
                else:
                    self.input_height,self.input_width=shape[:2]
            elif self.input_cropping or self.target_cropping or self.float_cropping:
                raise ValueError(DIMS_REQUIRED_ERROR)


    def _return_data(self,im,profile,return_profile):
        if self.augment:
            im=proc.augment(im,self.k,self.flip)
        if return_profile:
            return im, profile
        else:
            return im


#
# HELPERS
#
def process_input(
        im,
        rotate=False,
        flip=False,     
        input_bands=None,
        band_indices=None,
        indices_dict=None,
        padding=None,
        padding_value=0,
        cropping=None,
        bounds=None,
        means=None,
        stdevs=None,
        preprocess=None,
        dtype=INPUT_DTYPE):
    if preprocess:
        im=preprocess(im)
    im=proc.augment(im,k=rotate,flip=flip)
    if cropping:
        im=proc.crop(im,cropping)
    if band_indices:
        index_bands=[indices.index(im,idx,indices_dict) for idx in band_indices]
    if means is not None:
        if stdevs is None:
            im=proc.center(im,means=means,to_int=False)
        else:
            im=proc.normalize(im,means=means,stdevs=stdevs)
    if input_bands:
        if BANDS_FIRST:
            im=im[input_bands]
        else:
            im=im[:,:,input_bands]
    if band_indices:
        if BANDS_FIRST:
            if input_bands is False:
                im=np.vstack([index_bands])
            else:
                im=np.vstack([im,index_bands])
        else:
            if input_bands is False:
                im=np.dstack(index_bands)
            else:
                im=np.dstack([im]+index_bands)
    if (not cropping) and padding:
        im=proc.pad(im,padding=padding,value=padding_value)
    if bounds:
        for i,b in bounds.items():
            i=int(i)
            im[i]=im[i].clip(min=b.get('min'),max=b.get('max'))
    return im.astype(dtype)


def process_target(
        im,
        rotate=False,
        flip=False,
        value_map=None,
        list_value_map=False,
        default_mapped_value=proc.DEFAULT_VMAP_VALUE,
        categorical=False,
        nb_categories=None,
        padding=None,
        padding_value=0,
        cropping=None,
        expand_axis=None,
        preprocess=None,
        squeeze=True,
        dtype=TARGET_DTYPE):
    if preprocess:
        im=preprocess(im)
    im=proc.augment(im,k=rotate,flip=flip)
    if squeeze:
        im=np.squeeze(im)
    if value_map and list_value_map:
        out=[]
        for vmap in value_map:
            out.append(_post_process_target_images(
                im,
                vmap,
                categorical,
                nb_categories,
                cropping,
                padding,
                padding_value,
                expand_axis,
                dtype))
    else:
        out=_post_process_target_images(
                im,
                value_map,
                categorical,
                nb_categories,
                cropping,
                padding,
                padding_value,
                expand_axis,
                dtype)
    return out



def _post_process_target_images(
        im,
        value_map,
        categorical,
        nb_categories,
        cropping,
        padding,
        padding_value,
        expand_axis,
        dtype):
    if value_map:
        im=proc.map_values(im,value_map)
    if categorical:
        im=proc.to_categorical(im,nb_categories)
    if cropping:
        im=proc.crop(im,cropping)
    elif padding:
        im=proc.pad(im,padding=padding,value=padding_value)
    if expand_axis is not None:
        if expand_axis is True:
            expand_axis=0
        im=np.expand_dims(im,axis=expand_axis)
    return im.astype(dtype)




#
# TILLER
#
class Tiller(object):
    """ Tiller
    
    For a given boundary shape generate windows (x-offset, y-offset, width, height) of a given size and overlap

    Usage:
        im=np.arange(1024**2).reshape((1024,1024))
        tiller=hand.Tiller(boundary_shape=im.shape,size=100,overlap=10)
        xoff,yoff,width,height=tiller[0]
        print("NB WINDOWS:",len(tiller))
        print("WINDOW-0:",xoff,yoff,width,height)
        im[yoff:yoff+height,xoff:xoff+width]
        ### output:
        NB WINDOWS: 115600
        WINDOW-0: 1 1 5 5
        array([[1025, 1026, 1027, 1028, 1029],
               [2049, 2050, 2051, 2052, 2053],
               [3073, 3074, 3075, 3076, 3077],
               [4097, 4098, 4099, 4100, 4101],
               [5121, 5122, 5123, 5124, 5125]])

    Args:
        boundary_width/height<int|None>: 
            - width/height of boundary
            - required if boundary shape not specified
        boundary_shape<tuple|None>:
            - shape tuple
            - required if width/height not specified
        size<int>: tile size (width/height - only supports square tiles)
        overlap<int>: overlap between tiles
    """
    def __init__(
            self,
            boundary_width=None,
            boundary_height=None,
            boundary_shape=None,
            size=DEFAULT_SIZE,
            overlap=DEFAULT_OVERLAP):
        if not overlap:
            overlap=0
        self.size=size
        self.inner_size=self.size-2*overlap
        self._set_shape_attributes(
            boundary_width,
            boundary_height,
            boundary_shape,
            overlap)
        self.length=self.cols*self.rows
            
            
    def column_row(self,index):
        col=index//self.rows
        row=index-col*self.rows
        return int(col), int(row)
    

    def window(self,index=None,col=None,row=None):
        if index is not None:
            col,row=self.column_row(index)
        self.col,self.row=col,row  
        col_off=self.col_offset+col*self.size
        row_off=self.row_offset+row*self.size
        return (col_off,row_off,self.size,self.size)
    

    def __len__(self):
        return self.length
            

    def __getitem__(self, index):
        if (index>=self.length):
            raise IndexError
        self.index=index
        return self.window(index=self.index)
            
            
    #
    # INTERNEL
    #
    def _set_shape_attributes(
            self,
            boundary_width,
            boundary_height,
            boundary_shape,
            overlap):
        if boundary_shape:
            boundary_height,boundary_width=boundary_shape
        self.cols=math.floor((boundary_width-2*overlap)/self.inner_size)
        self.rows=math.floor((boundary_height-2*overlap)/self.inner_size)
        self.width=2*overlap+self.cols*self.inner_size
        self.height=2*overlap+self.rows*self.inner_size
        self.col_offset=math.floor((boundary_width-self.width)/2)
        self.row_offset=math.floor((boundary_height-self.height)/2)
        self.overlap=overlap

